#ifndef __ASM_OFFSETS_H__
#define __ASM_OFFSETS_H__
#define TASK_STATE 0
#define TASK_FLAGS 8
#define TASK_SIGPENDING 16
#define TASK_ADDR_LIMIT 24
#define TASK_EXEC_DOMAIN 32
#define TASK_NEED_RESCHED 40
#define TASK_SIZE 1096
#define STACK_SIZE 16384
#define HAE_CACHE 0
#define HAE_REG 8
#endif /* __ASM_OFFSETS_H__ */
