/*
 * linux/include/asm-arm/arch-arc/a.out.h
 *
 * Copyright (C) 1996 Russell King
 *
 * Acorn Archimedes/A5000 a.out.h specs
 */
#ifndef __ASM_ARCH_A_OUT_H
#define __ASM_ARCH_A_OUT_H

#ifdef __KERNEL__
#define STACK_TOP		(0x01a00000)
#define LIBRARY_START_TEXT	(0x00c00000)
#endif

#endif

